/*@/// doc */
// Hashing of 32 bit integers
//
// (c) 2013..2017 by Jasper L. Neumann
// www.sirrida.de / programming.sirrida.de
// E-Mail: info@sirrida.de
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
// First version: 2013-11
// Last change: 2014-02
//
// Based on the minimal hashing code generator of
// Robert John Jenkins Junior (Bob Jenkins)
// http://burtleburtle.net/bob/hash/perfect.html

// I corrected some errors and tweaked code and parameters to speed it up.
// String handling and special treatment of small switch sets are left out.
// tabb containing only zeros is now optimized away.
// I added reversible and simple hashing as well as 2 new a/b hashing methods.
// Now it is a reusable library.

// Bugs corrected:
// 1. tabh was too small for minimal hashing => AV in apply().
// 2. For 17 keys, slow mode, non-minimal the code generator could choose
//    gen_ab_2 (perfhex.c, hexn(), case 2) and the shift right amount for b
//    could get 32. This does not always work as expected since at least for
//    x86 using 32 bit registers such a shift does *nothing* and the resulting
//    code could trap on accessing invalid tabb elements.
// 3. For slen>131072 alen got an undefined value => AV / heap overflow.
// 4. For fast minimal hashing alen was probably guessed smaller than intended
//    (quirk, no real impact).

/*
Here are some notes from Bob Jenkins' sources:

perfect.c: code to generate code for a hash for perfect hashing.
(c) Bob Jenkins, September 1996, December 1999
You may use this code in any way you wish, and it is free.  No warranty.
I hereby place this in the public domain.
Source is http://burtleburtle.net/bob/c/perfect.c

This generates a minimal perfect hash function.  That means, given a
set of n keys, this determines a hash function that maps each of
those keys into a value in 0..n-1 with no collisions.

The perfect hash function first uses a normal hash function on the key
to determine (a,b) such that the pair (a,b) is distinct for all
keys, then it computes a^scramble[tab[b]] to get the final perfect hash.
tab[] is an array of 1-byte values and scramble[] is a 256-term array of
2-byte or 4-byte values.  If there are n keys, the length of tab[] is a
power of two between n/3 and n.

I found the idea of computing distinct (a,b) values in "Practical minimal
perfect hash functions for large databases", Fox, Heath, Chen, and Daoud,
Communications of the ACM, January 1992.  They found the idea in Chichelli
(CACM Jan 1980).  Beyond that, our methods differ.

The key is hashed to a pair (a,b) where a in 0..*alen*-1 and b in
0..*blen*-1.  A fast hash function determines both a and b
simultaneously.  Any decent hash function is likely to produce
hashes so that (a,b) is distinct for all pairs.  I try the hash
using different values of *salt* until all pairs are distinct.

The final hash is (a XOR scramble[tab[b]]).  *scramble* is a
predetermined mapping of 0..255 into 0..smax-1.  *tab* is an
array that we fill in in such a way as to make the hash perfect.

First we fill in all values of *tab* that are used by more than one
key.  We try all possible values for each position until one works.

This leaves m unmapped keys and m values that something could hash to.
If you treat unmapped keys as lefthand nodes and unused hash values
as righthand nodes, and draw a line connecting each key to each hash
value it could map to, you get a bipartite graph.  We attempt to
find a perfect matching in this graph.  If we succeed, we have
determined a perfect hash for the whole set of keys.

*scramble* is used because (a^tab[i]) clusters keys around *a*.
*/
/*@\\\0000005001+00CE*/

// TODO: Cost function with options
// TODO: Treat also 64 bit integers and strings?

#include "hashlib.hpp"

/*@/// tools */
/*@/// static mycall void break_point() */
static mycall void break_point(
  ) {
  // cout << "break_point\n";
  // exit(1);  /////
  }
/*@\\\+0903*/

/*@/// static mycall t_32u min_32(t_32u a, t_32u b) */
static mycall t_32u min_32(
  t_32u a,
  t_32u b
  ) {

  if (a < b)
    return a;
  else
    return b;
  }
/*@\\\0000000109+9FB9*/
/*@/// static mycall t_32u ror_32(t_32u x, t_32u shift) */
static mycall t_32u ror_32(
  t_32u x,
  t_32u shift
  ) {

  return (x >> shift) | (x << (32-shift));
  }
/*@\\\0000000201+10C3*/
/*@/// static mycall t_32u rol_32(t_32u x, t_32u shift) */
static mycall t_32u rol_32(
  t_32u x,
  t_32u shift
  ) {

  return (x << shift) | (x >> (32-shift));
  }
/*@\\\+2B4D*/
/*@/// static mycall t_32u gcd(t_32u a, t_32u b) */
static mycall t_32u gcd(
  t_32u a,
  t_32u b
  ) {
// Greatest common divisor

  t_32u t;

  while (b != 0) {
    t = a % b;
    a = b;
    b = t;
    }
  return a;
  }
/*@\\\000C000901000E0100010F+575F*/
/*@/// static mycall t_32u mul_inv_32(t_32u x) */
static mycall t_32u mul_inv_32(
  t_32u x
  ) {

  t_32u xn,t;

  if ((x & 1) == 0)
    return 0;
  else {
    xn = x;
    while (true) {
      t = x*xn;
      if (t == 1)
        break;
      xn = xn*(2-t);
      }
    return xn;
    }
  }
/*@\\\+F413*/
/*@/// static mycall t_32u floor_log2(t_32u x) */
static mycall t_32u floor_log2(
  t_32u x
  ) {

  t_32u res;

  res = (t_32u)(-1);
  while (x != 0) {
    ++res;
    x >>= 1;
    }
  return res;
  }
/*@\\\+256A*/
/*@/// static mycall t_32u ceil_log2(t_32u x) */
static mycall t_32u ceil_log2(
  t_32u x
  ) {

  switch (x) {
    case 0:
      break_point();
      return (t_32u)(-1);
    case 1:
      return 0;
    default:
      return floor_log2(x-1)+1;
    }
  }


//static mycall t_32u ceil_log2(
//  t_32u val
//  ) {
//// return the ceiling of the log (base 2) of val
//
//  t_32u i;
//
//  i = 0;
//  while ((t_32u)(1) << i < val)
//    ++i;
//  return i;
//  }
/*@\\\000000170B+3595*/
/*@\\\0030000601000601000601+8B49*/

/*@/// define */
#define UB4BITS 32

// test_simple_perfect
#define MAX_SET 256
  // Max. size of hash table for simple hashing
#define SEARCH_START 0x04d7651f
  // Hacker's Delight, de Bruijn, fig.5-26, maps 2**x
#define SEARCH_OFFSET 0x61c88647
  // (1-1/((1+sqrt 5)/2)) * 0x1_0000_0000
#define ZERO_MEM(var,size) memset(&var,0,(size))
#define ZERO_OUT(var) memset(&var,0,sizeof(var))

#ifndef TRACE
  #define TRACE(s)
  // Some general logging; string s; num(t_32u)=>string must be defined
#endif
#ifndef TRACE_ODO
  #define TRACE_ODO(s)
  // Some counting logging; string s; num(t_32u)=>string must be defined
#endif
#ifndef RAISE
  #define RAISE(s)
  // Might be used to let self_test() raise exceptions; string s
#endif
/*@\\\0000000E10+A7BB*/

/*@/// static mycall t_32u permute(...) */
static mycall t_32u permute(
  t_32u x,                                  // input, a value in some range
  t_32u nbits                             // input, number of bits in range
  ) {
// compute p(x), where p is a permutation of 0..(1<<nbits)-1
// permute(0)=0.  This is intended and useful.

  t_int i;
  t_32u mask;
  t_int const2;
  t_int const3;
  t_int const4;
  t_int const5;

  mask   = ((t_32u)(1) << nbits)-1;  // all ones
  const2 = 1 + nbits/2;
  const3 = 1 + nbits/3;
  const4 = 1 + nbits/4;
  const5 = 1 + nbits/5;
  for (i = 0; i <= 20-1; ++i) {
    x = (x + (x << const2)) & mask;
    x = (x ^ (x >> const3));
    x = (x + (x << const4)) & mask;
    x = (x ^ (x >> const5));
    }
  return x;
  }
/*@\\\+E28C*/

/*@/// static mycall t_32u calc_hash_tab(const tr_gen_opt &opt, const tr_key &key) */
static mycall t_32u calc_hash_tab(
  const tr_gen_opt &opt,
  const tr_key &key
  ) {

  if (opt.tabb_len == 0)
    return key.a_k;  // OPTIMIZE: do not calculate b
  else if (opt.tabb_len == 1)
    return key.a_k ^ opt.p_tabb[0];
  // OPTIMIZE: do not calculate b, tabb[0]=const!
  else if (opt.scramble_len == 0)
    return key.a_k ^ opt.p_tabb[key.b_k];
  else
    return key.a_k ^ opt.p_scramble[opt.p_tabb[key.b_k]];
  }
/*@\\\0000000C01+6375*/
/*@/// static mycall void calc_ab_*(const tr_gen_opt &opt, tr_key &key) */
/*@/// static mycall void calc_ab_reversible(const tr_gen_opt &opt, tr_key &key) */
static mycall void calc_ab_reversible(
  const tr_gen_opt &opt,
  tr_key &key
  ) {

  t_32u x;

  x = key.key_k;
  key.a_k = ror_32(x-opt.salt, opt.a_shr) * opt.a_mask;
  }
/*@\\\+8C46*/

/*@/// static mycall void calc_ab_and(const tr_gen_opt &opt, tr_key &key) */
static mycall void calc_ab_and(
  const tr_gen_opt &opt,
  tr_key &key
  ) {

  t_32u x;

  x = key.key_k;
  key.a_k = x & opt.a_mask;
  }
/*@\\\000000050B+900E*/
/*@/// static mycall void calc_ab_shr(const tr_gen_opt &opt, tr_key &key) */
static mycall void calc_ab_shr(
  const tr_gen_opt &opt,
  tr_key &key
  ) {

  t_32u x;

  x = key.key_k;
  key.a_k = x >> opt.a_shr;
  }
/*@\\\+5098*/
/*@/// static mycall void calc_ab_rol(const tr_gen_opt &opt, tr_key &key) */
static mycall void calc_ab_rol(
  const tr_gen_opt &opt,
  tr_key &key
  ) {

  t_32u x;

  x = key.key_k;
  key.a_k = rol_32(x, opt.salt) & opt.a_mask;
  }
/*@\\\+4E20*/
/*@/// static mycall void calc_ab_rol_xor(const tr_gen_opt &opt, tr_key &key) */
static mycall void calc_ab_rol_xor(
  const tr_gen_opt &opt,
  tr_key &key
  ) {

  t_32u x;

  x = key.key_k;
  key.a_k = (rol_32(x, opt.salt) ^ x) & opt.a_mask;
  }
/*@\\\0000000621+7218*/
/*@/// static mycall void calc_ab_rol_add(const tr_gen_opt &opt, tr_key &key) */
static mycall void calc_ab_rol_add(
  const tr_gen_opt &opt,
  tr_key &key
  ) {

  t_32u x;

  x = key.key_k;
  key.a_k = (rol_32(x, opt.salt) + x) & opt.a_mask;
  }
/*@\\\0000000621+3440*/
/*@/// static mycall void calc_ab_rol_sub(const tr_gen_opt &opt, tr_key &key) */
static mycall void calc_ab_rol_sub(
  const tr_gen_opt &opt,
  tr_key &key
  ) {

  t_32u x;

  x = key.key_k;
  key.a_k = (rol_32(x, opt.salt) - x) & opt.a_mask;
  }
/*@\\\0000000620+4349*/
/*@/// static mycall void calc_ab_mul(const tr_gen_opt &opt, tr_key &key) */
static mycall void calc_ab_mul(
  const tr_gen_opt &opt,
  tr_key &key
  ) {

  t_32u x;

  x = key.key_k;
  key.a_k = (t_32u)(x*opt.salt) >> opt.a_shr;
  }
/*@\\\+9161*/

/*@/// static mycall void calc_ab_1(const tr_gen_opt &opt, tr_key &key) */
static mycall void calc_ab_1(
  const tr_gen_opt &opt,
  tr_key &key
  ) {

  t_32u x;

  x = key.key_k;
  // if (opt.a_shl == opt.a_shr)
  //   break_point();  // can happen; optimize as key.a_k = x & opt.a_mask;
                       // in this case also opt.b_shr=0
  if (opt.a_shr == 0)
    break_point();  // TCH
  if (opt.a_shr == 32) {
    break_point();
    key.a_k = 0;  // TCH
    }
  else
    key.a_k = (x << opt.a_shl) >> opt.a_shr;

  // if (opt.b_shr == 0)
  //   break_point();  // can happen
  if (opt.b_shr == 32) {
    key.b_k = 0;  // CAN HAPPEN (don't care, opt.b_mask should be 0)
    if (opt.b_mask != 0)
      break_point();  // TCH
    }
  else
    key.b_k = (x >> opt.b_shr) & opt.b_mask;
  }
/*@\\\0000000A13+5B0C*/
/*@/// static mycall void calc_ab_2(const tr_gen_opt &opt, tr_key &key) */
static mycall void calc_ab_2(
  const tr_gen_opt &opt,
  tr_key &key
  ) {

  t_32u x;

  x = key.key_k;
  // if (opt.a_shr == 0)
  //   break_point();  // can happen
  key.a_k = (x >> opt.a_shr) & opt.a_mask;

  if (opt.b_shl == opt.b_shr)
    break_point();  // TCH
  if (opt.b_shr == 0)
    break_point();  // TCH
  if (opt.b_shr == 32)  // BUG 2 in Jenkin's code (not caught)
    key.b_k = 0;  // CAN HAPPEN, optimized away for app by finalize_perfect
  else
    key.b_k = (x << opt.b_shl) >> opt.b_shr;
  }
/*@\\\+422D*/
/*@/// static mycall void calc_ab_3(const tr_gen_opt &opt, tr_key &key) */
static mycall void calc_ab_3(
  const tr_gen_opt &opt,
  tr_key &key
  ) {

  t_32u x;

  x = key.key_k;
  x += opt.salt;
  if (opt.do_xor_shr_16)
    x ^= x >> 16;
  if (opt.do_add_shl_8)
    x += x << 8;
  x ^= x >> 4;

  if (opt.a_shr == 0)
    break_point();
  if (opt.a_shr == 32) {
    break_point();
    key.a_k = 0;  // TCH
    }
  else if (opt.a_shl == 0)
    key.a_k = x >> opt.a_shr;  // extra case, not identical to below
  else
    key.a_k = ((x << opt.a_shl) + x) >> opt.a_shr;

  // if (opt.b_shr == 0)
  //   break_point();  // can happen
  if (opt.b_shr == 32) {
    key.b_k = 0;  // CAN HAPPEN (don't care, opt.b_mask should be 0)
    if (opt.b_mask != 0)
      break_point();  // TCH
    }
  else
    key.b_k = (x >> opt.b_shr) & opt.b_mask;
  }
/*@\\\+294E*/

/*@/// static mycall void calc_ab_x1(const tr_gen_opt &opt, tr_key &key) */
static mycall void calc_ab_x1(
  const tr_gen_opt &opt,
  tr_key &key
  ) {
// 32x32=>32 mul

  t_32u tmp;
  t_32u x;

  x = key.key_k;

  tmp = x * opt.salt;

  key.a_k = (tmp >> opt.a_shr) & opt.a_mask;

  if (opt.b_shr == 32)
    key.b_k = 0;  // CAN HAPPEN
  else
    key.b_k = tmp >> opt.b_shr;
  }
/*@\\\+2D44*/
/*@/// static mycall void calc_ab_x3(const tr_gen_opt &opt, tr_key &key) */
static mycall void calc_ab_x3(
  const tr_gen_opt &opt,
  tr_key &key
  ) {
// 32x32=>64 mul

  t_64u tmp;
  t_32u x;

  x = key.key_k;

  tmp = (t_64u)(x) * (t_64u)(opt.salt);

  key.a_k = (t_32u)(tmp >> 32) & opt.a_mask;

  if (opt.b_shr == 32)
    key.b_k = 0;  // CAN HAPPEN
  else
    key.b_k = (t_32u)(tmp) >> opt.b_shr;
  }
/*@\\\+FB86*/
// other ideas:
// tmp:=x*salt; b:=tmp>>b_shr; a:=((tmp+x)>>a_shr)&a_mask
// tmp:=x*salt; b:=tmp>>b_shr; a:=((tmp>>a_shr)+x)&a_mask
/*@\\\0000000105+0925*/

/*@/// static const tf_calc_ab a_f_calc_ab[HM_MAX]={...}; */
static const tf_calc_ab a_f_calc_ab[HM_MAX]={
  nil,
  calc_ab_reversible,
  calc_ab_and,
  calc_ab_shr,
  calc_ab_rol,
  calc_ab_rol_xor,
  calc_ab_rol_add,
  calc_ab_rol_sub,
  calc_ab_mul,
  calc_ab_1,
  calc_ab_2,
  calc_ab_3,
  calc_ab_x1,
  nil,  // calc_ab_x2,
  calc_ab_x3
  };
/*@\\\+B33B*/

/*@/// class to_perfect_hash */
/*@/// void to_perfect_hash::init() */
void to_perfect_hash::init(
  ) {

  /*@/// zero all */
  ZERO_OUT(this->priv);
  ZERO_OUT(this->state_engine);

  ZERO_OUT(this->calc_opt);
  ZERO_OUT(this->gen_opt);

  ZERO_OUT(this->cb);
  /*@\\\0000000109+5ED9*/

  this->defaults();
  this->clean();
  }
/*@\\\0000000401+DFF9*/
/*@/// void to_perfect_hash::defaults() */
void to_perfect_hash::defaults(
  ) {

  // Limits, usable up to ~700_000 elements
  this->calc_opt.minimal = false;
  this->calc_opt.fast = false;
  this->calc_opt.fast_trans = true;  // true appears to be faster
  this->calc_opt.nr_mul_checks = 100000;
  this->calc_opt.listlen_limit = 100;
  this->calc_opt.retry_initkey = 2048;
  this->calc_opt.retry_perfect = 10;  // 200;
  this->calc_opt.count_limit = 100000;  // max. work per element, should be const
  this->calc_opt.work_limit = 0;  // max sum work, should scale with sqrt(nkeys)
  // this->calc_opt.work_limit = 10000000;  // max sum work, should scale with sqrt(nkeys)
  this->calc_opt.use_scramble = 4096;
  this->calc_opt.near_minimal_factor = 4.0;  // way too large to be effective
  this->calc_opt.min_load_factor = 0.4;
  this->calc_opt.keyspace_factor = 0.85;
  }
/*@\\\0000000742+F5B5*/
/*@/// void to_perfect_hash::stop() */
void to_perfect_hash::stop(
  ) {

  this->clean();
  }
/*@\\\+B2AB*/
/*@/// void to_perfect_hash::add_key(t_32u value, t_pointer user_k) */
void to_perfect_hash::add_key(
  t_32u value,
  t_pointer user_k
  ) {

  tpr_key mykey;

  mykey = (tpr_key)calloc(1, sizeof(*mykey));

  mykey->key_k = value;
  mykey->user_k = user_k;

  mykey->next_k = this->priv.keys;
  this->priv.keys = mykey;
  ++this->priv.nkeys;
  }
/*@\\\000000051E+1BB1*/
/*@/// t_bool to_perfect_hash::generate_hash() */
t_bool to_perfect_hash::generate_hash(
  ) {

  if (this->calc_opt.min_load_factor > 1.0)
    this->calc_opt.min_load_factor = 1.0;
  if (this->calc_opt.near_minimal_factor < 1.0)
    this->calc_opt.near_minimal_factor = 1.0;
  if (this->priv.p_inv_hash != nil) {
    free(this->priv.p_inv_hash);
    this->priv.p_inv_hash = nil;
    }

  if (this->priv.nkeys <= 1) {
    // 0/1 elements
    return false;
    }
  else if (this->test_reversible())
    return true;
  else if (this->test_simple_perfect())
    return true;
  else if (this->findhash())
    return true;
  else
    return false;
  }
/*@\\\0000000119+6E80*/
/*@/// void to_perfect_hash::do_generate(t_pointer gen_stuff) */
void to_perfect_hash::do_generate(
  t_pointer gen_stuff
  ) {

  if (assigned(this->cb.a_gen[this->gen_opt.hash_method])) {
      this->cb.a_gen[this->gen_opt.hash_method](this->gen_opt, gen_stuff);
    }
  }
/*@\\\0000000325+8D6F*/
/*@/// t_32u to_perfect_hash::calc_hash(t_32u key_k) const */
t_32u to_perfect_hash::calc_hash(
  t_32u key_k
  ) const {

  tr_key my_key;

  my_key.key_k = key_k;
  a_f_calc_ab[this->gen_opt.hash_method](this->gen_opt, my_key);
  return calc_hash_tab(this->gen_opt, my_key);
  }
/*@\\\0000000601+5DDD*/
/*@/// void to_perfect_hash::step_keys(tf_step_key f_step_key, t_pointer step_stuff) const */
void to_perfect_hash::step_keys(
  tf_step_key f_step_key,
  t_pointer step_stuff
  ) const {

  tpr_key mykey;

  if (assigned(f_step_key)) {
    mykey = this->priv.keys;
    while (mykey != nil) {
      f_step_key(mykey->key_k, mykey->user_k, step_stuff);
      mykey = mykey->next_k;
      }
    }
  }
/*@\\\+CD1C*/
/*@/// t_bool to_perfect_hash::self_test() const */
t_bool to_perfect_hash::self_test(
  ) const {
// to be used before add_key or after generate_hash

  tr_key my_key;
  tpr_key p_key;
  t_bool *hash_target;  // array
  t_32u n;
  t_32u hash;
  t_32u max_hash;
  t_32u value;
  t_pointer key_stuff;
  t_bool res;
  tf_calc_ab f_calc_ab;

  f_calc_ab = a_f_calc_ab[this->gen_opt.hash_method];
  if (this->gen_opt.hash_used_max > this->gen_opt.hash_gen_max) {
    break_point();
    RAISE("hash_used_max>hash_gen_max");
    return false;
    }

  if (this->priv.nkeys <= 1) {
    // 0/1 elements => no check
    return true;
    }

  ZERO_OUT(my_key);

  res = false;
  key_stuff = nil;  // not used
  hash_target = nil;
  // try
    hash_target = (t_bool*)
      calloc(this->gen_opt.hash_used_max+1, sizeof(hash_target[0]));
    max_hash = 0;
    n = 0;
    p_key = this->priv.keys;
    while (p_key != nil) {
      ++n;
      my_key.key_k = p_key->key_k;
      f_calc_ab(this->gen_opt, my_key);
      if ((my_key.b_k > this->gen_opt.tabb_len) &&
         (this->gen_opt.tabb_len != 0)) {
        break_point();
        RAISE("b_k too large");
        res = false;
        goto STOP;
        }
      hash = calc_hash_tab(this->gen_opt, my_key);
      if (hash > max_hash)
        max_hash = hash;
      if (hash > this->gen_opt.hash_used_max) {
        break_point();
        RAISE("hash>hash_used_max");
        res = false;
        goto STOP;
        }
      if (hash_target[hash]) {
        break_point();
        RAISE("hash dupe");
        res = false;
        goto STOP;
        }
      hash_target[hash] = true;

      value = 171161;  // dummy value
      if (! this->inv_hash(hash, value, key_stuff)) {
        break_point();
        RAISE("absent inverse hash");
        res = false;
        goto STOP;
        }
      if (value != p_key->key_k) {
        break_point();
        RAISE("invalid inverse hash");
        res = false;
        goto STOP;
        }

      p_key = p_key->next_k;
      }

    if (n != this->priv.nkeys) {
      break_point();
      RAISE("nkeys defective");
      res = false;
      goto STOP;
      }
    if (max_hash != this->gen_opt.hash_used_max) {
      break_point();
      RAISE("max hash != hash_used_max");
      res = false;
      goto STOP;
      }
    res = true;
  // finally
STOP:
    if (hash_target != nil)
      free(hash_target);

  return res;
  }
/*@\\\0000002601+2A4A*/
/*@/// void to_perfect_hash::clean() */
void to_perfect_hash::clean(
  ) {

  tpr_key mykey,next;

  mykey = this->priv.keys;
  while (mykey != nil) {
    next = mykey->next_k;
    free(mykey);
    mykey = next;
    }
  if (this->priv.p_tabb != nil)
    free(this->priv.p_tabb);
  if (this->gen_opt.p_tabb != nil)
    free(this->gen_opt.p_tabb);
  if (this->priv.p_scramble != nil)
    free(this->priv.p_scramble);
  if (this->priv.p_inv_hash != nil)
    free(this->priv.p_inv_hash);
  ZERO_OUT(this->state_engine);
  ZERO_OUT(this->priv);
  ZERO_OUT(this->gen_opt);
  }
/*@\\\+5210*/
/*@/// t_bool to_perfect_hash::inv_hash(t_32u hash, t_32u &key, t_pointer &key_stuff) const */
t_bool to_perfect_hash::inv_hash(
  t_32u hash,
  t_32u &key,
  t_pointer &key_stuff
  ) const {

  tpr_key mykey;

  // Don't touch result parameters if there is no matching key
  if ( (this->priv.p_inv_hash != nil) &&
       (hash <= this->gen_opt.hash_used_max) ) {
    mykey = this->priv.p_inv_hash[hash].key_h;
    if (mykey != nil) {
      key = mykey->key_k;
      key_stuff = mykey->user_k;
      return true;
      }
    }
  return false;
  }
/*@\\\+4588*/

/*@/// mycall void to_perfect_hash::calc_rev_cfg(tr_reversible &rev, t_32u min) */
mycall void to_perfect_hash::calc_rev_cfg(tr_reversible &rev, t_32u min) {

  tpr_key mykey;
  t_32u key;
  t_32u q;
  t_32u v;

  /*@/// gcd = */
  rev.gcd = 0;
  mykey = this->priv.keys;
  while (mykey != nil) {
    key = mykey->key_k-min;
    if (key != 0) {
      if (rev.gcd == 0)
        rev.gcd = key;
      else
        rev.gcd = gcd(rev.gcd, key);
      }
    mykey = mykey->next_k;
    }
  /*@\\\+DC53*/
  /*@/// rev.ror,rev.inv = */
  rev.ror = 0;
  rev.inv = 0;
  if (rev.gcd != 0) {
    q = rev.gcd;
    while ((q & 1) == 0) {
      ++rev.ror;
      q = q >> 1;
      }
    rev.inv = mul_inv_32(q);
    }
  /*@\\\+D247*/
  /*@/// rev.max = */
  rev.max = 0;
  mykey = this->priv.keys;
  while (mykey != nil) {
    v = ror_32(mykey->key_k-min, rev.ror) * rev.inv;
    if (v > rev.max)
      rev.max = v;
    mykey = mykey->next_k;
    }
  /*@\\\+4EBE*/
  }
/*@\\\0000000A01+F815*/
/*@/// mycall t_bool to_perfect_hash::test_reversible() */
mycall t_bool to_perfect_hash::test_reversible(
  ) {

  tpr_key mykey;
  t_32u key;
  t_32u min_u;
  t_32s min_s;
  tr_reversible rev_u;
  tr_reversible rev_s;

  if (! assigned(this->cb.a_gen[HM_REVERSIBLE])) {
    return false;
    }

  /*@/// min_*,max_* = */
  mykey = this->priv.keys;
  key = mykey->key_k;
  min_u = key;
  min_s = (t_32s)key;
  while (mykey != nil) {
    key = mykey->key_k;
    if (min_u > key)
      min_u = key;
    if (min_s > (t_32s)key)
      min_s = (t_32s)key;
    mykey = mykey->next_k;
    }
  /*@\\\+BD75*/
  calc_rev_cfg(rev_u, min_u);
  calc_rev_cfg(rev_s, min_s);

  if (rev_s.max < rev_u.max) {
    // The values seem to be signed; optimize for that
    rev_u = rev_s;
    min_u = min_s;
    }

  if (rev_u.max >= 0x10000000) {
    // forget it, way too large
    return false;
    }

  // density = (t_float)this->priv.nkeys / (t_float)(rev_u.max+1);
  if (this->calc_opt.minimal) {
    if (rev_u.max+1 != this->priv.nkeys)  // (density != 1.0)
      return false;
    }
  else {
    if ((t_float)this->priv.nkeys / (t_float)(rev_u.max+1)  // density
        < this->calc_opt.min_load_factor)
      return false;
    }

  this->gen_opt.hash_method = HM_REVERSIBLE;
  this->gen_opt.salt = min_u;
  this->gen_opt.a_shr = rev_u.ror;
  this->gen_opt.a_mask = rev_u.inv;
  this->gen_opt.hash_used_max = rev_u.max;  // also filled by calc_max()
  this->gen_opt.hash_gen_max = (t_32u)(-1);
  this->gen_opt.check_value = (rev_u.max+1 != this->priv.nkeys);
  this->gen_opt.check_range = true;
  this->calc_max();  // fill inverse hash table
  return true;
  }
/*@\\\0000001901+26FE*/

/*@/// mycall t_bool to_perfect_hash::check_direct(...) */
mycall t_bool to_perfect_hash::check_direct(
  tq_hash_method hash_method,
  t_32u x
  ) {
/*@/// var */

  tpr_key mykey;
  t_bool s[MAX_SET];
  t_32u v;
  tf_calc_ab f_calc_ab;
  tf_generate f_gen;
/*@\\\0000000313+40D1*/

  f_calc_ab = a_f_calc_ab[hash_method];
  f_gen = this->cb.a_gen[hash_method];
  if (! assigned(f_gen)) {
    return false;
    }

  this->gen_opt.salt = x;
  ZERO_MEM(s[0], MAX_SET*sizeof(s[0]));

  /*@/// loop */
  mykey = this->priv.keys;
  while (mykey != nil) {
    f_calc_ab(this->gen_opt, *mykey);
    v = mykey->a_k;
    if (s[v]) {
      return false;
      }
    s[v] = true;
    mykey = mykey->next_k;
    }
  /*@\\\+9122*/

  this->gen_opt.hash_method = hash_method;
  return true;
  }
/*@\\\0000000701+0BBE*/
/*@/// mycall t_bool to_perfect_hash::test_simple_perfect() */
mycall t_bool to_perfect_hash::test_simple_perfect(
  ) {

  t_bool found;
  t_32u mul;
  t_32u i;
  t_32u bits;

  found = false;
  bits = ceil_log2(this->priv.nkeys);
  while (true) {
    this->priv.alen = (t_32u)(1) << bits;
    if (this->priv.alen > MAX_SET) {  // set operation
      goto STOP;  // false; break while true
      }
    if ((t_float)this->priv.nkeys / (t_float)this->priv.alen
        < this->calc_opt.min_load_factor) {
      goto STOP;  // false; break while true
      }
    found = true;
    this->gen_opt.a_shr = UB4BITS-bits;
    this->gen_opt.a_mask = this->priv.alen-1;
    // this->gen_opt.hash_used_max => below
    this->gen_opt.hash_gen_max = this->priv.alen-1;
    this->gen_opt.check_value = true;
    this->gen_opt.check_range = false;
    /*@/// if (this->check_direct(...))  goto STOP; */
    if (this->check_direct(HM_SIMPLE_AND, 0))
      goto STOP;  // true
    if (this->check_direct(HM_SIMPLE_SHR, 0))
      goto STOP;  // true
    if (assigned(this->cb.a_gen[HM_SIMPLE_ROL])) {
      for (i = 1; i <= 31; ++i) {
        if (this->check_direct(HM_SIMPLE_ROL, i))
          goto STOP;  // true
        }
      }
    if (assigned(this->cb.a_gen[HM_SIMPLE_ROL_XOR])) {
      for (i = 1; i <= 31; ++i) {
        if (this->check_direct(HM_SIMPLE_ROL_XOR, i))
          goto STOP;  // true
        }
      }
    if (assigned(this->cb.a_gen[HM_SIMPLE_ROL_ADD])) {
      for (i = 1; i <= 31; ++i) {
        if (this->check_direct(HM_SIMPLE_ROL_ADD, i))
          goto STOP;  // true
        }
      }
    if (assigned(this->cb.a_gen[HM_SIMPLE_ROL_SUB])) {
      for (i = 1; i <= 31; ++i) {
        if (this->check_direct(HM_SIMPLE_ROL_SUB, i))
          goto STOP;  // true
        }
      }

    if (assigned(this->cb.a_gen[HM_SIMPLE_MUL])) {
      mul = SEARCH_START;
      for (i = 1; i <= this->calc_opt.nr_mul_checks; ++i) {
        if (this->check_direct(HM_SIMPLE_MUL, mul))
          goto STOP;  // true
        mul += SEARCH_OFFSET;
        }
      }
    /*@\\\0000001827+6F98*/
    found = false;
    if (this->calc_opt.minimal) {
      goto STOP;  // false; break while true; no other tries
      }
    ++bits;
    }
STOP:

  if (found) {
    // this->gen_opt.hash_used_max => calc_max
    this->calc_max();
    }
  return found;
  }
/*@\\\+15ED*/

/*@/// mycall void to_perfect_hash::initalen() */
mycall void to_perfect_hash::initalen(
  ) {
/*
  returns
    alen, initial alen
    blen, initial blen
    slen, power of two greater or equal to max hash value
*/
// guess initial values for alen and blen
/*@/// doc */
/*
 * Find initial alen, blen
 * Initial alen and blen values were found empirically.  Some factors:
 *
 * If slen<256 there is no scramble, so tab[b] needs to cover 0..slen-1.
 *
 * alen and blen must be powers of 2 because the values in 0..alen-1 and
 * 0..blen-1 are produced by applying a bitmask to the initial hash function.
 *
 * alen must be less than slen, in fact less than nkeys, because otherwise
 * there would often be no i such that a^p_scramble[i] is in 0..nkeys-1 for
 * all the *a*s associated with a given *b*, so there would be no legal
 * value to assign to tab[b].  This only matters when we're doing a minimal
 * perfect hash.
 *
 * It takes around 800 trials to find distinct (a,b) with nkey=slen*(5/8)
 * and alen*blen = slen*slen/32.
 *
 * Values of blen less than slen/4 never work, and slen/2 always works.
 *
 * We want blen as small as possible because it is the number of bytes in
 * the huge array we must create for the perfect hash.
 *
 * When nkey <= slen*(5/8), blen=slen/4 works much more often with
 * alen=slen/8 than with alen=slen/4.  Above slen*(5/8), blen=slen/4
 * doesn't seem to care whether alen=slen/8 or alen=slen/4.  I think it
 * has something to do with 5/8 = 1/8 * 5.  For example examine 80000,
 * 85000, and 90000 keys with different values of alen.  This only matters
 * if we're doing a minimal perfect hash.
 *
 * When alen*blen <= 1<<UB4BITS, the initial hash must produce one integer.
 * Bigger than that it must produce two integers, which increases the
 * cost of the hash per character hashed.
 */
/*@\\\0000000507+00CE*/

  this->priv.slen = (t_32u)(1) << ceil_log2(this->priv.nkeys);
  // code moved from findhash

  if (this->calc_opt.minimal) {
    /*@/// minimal */
    switch (ceil_log2(this->priv.slen)) {
      /*@/// case 0: */
      case 0: {
        this->priv.alen = 1;
        this->priv.blen = 1;
        break;
        }
      /*@\\\0002000409000409+9376*/
      /*@/// case 1..8: */
      case 1:
      case 2:
      case 3:
      case 4:
      case 5:
      case 6:
      case 7:
      case 8: {
        // if (! this->calc_opt.minimal)  // always false!
        // BUG 4 in Jenkins' code
        if (this->calc_opt.fast)  // this was probably intended
          this->priv.alen = this->priv.slen;
        else
          this->priv.alen = this->priv.slen/2;
        this->priv.blen = this->priv.slen/2;
        break;
        }
      /*@\\\0000000B1A+D5FB*/
      /*@/// case 9..17: */
      case 9:
      case 10:
      case 11:
      case 12:
      case 13:
      case 14:
      case 15:
      case 16:
      case 17: {
        if (this->calc_opt.fast) {
          this->priv.alen = this->priv.slen/2;
          this->priv.blen = this->priv.slen/4;
          }
        else if (this->priv.slen/4 < this->calc_opt.use_scramble) {
          if (this->priv.nkeys <= this->priv.slen*0.52) {
            this->priv.alen = this->priv.slen/8;
            this->priv.blen = this->priv.slen/8;
            }
          else {
            this->priv.alen = this->priv.slen/4;
            this->priv.blen = this->priv.slen/4;
            }
          }
        else {
          if (this->priv.nkeys <= this->priv.slen*(5.0/8.0))
            this->priv.alen = this->priv.slen/8;
          else if (this->priv.nkeys <= this->priv.slen*(3.0/4.0))
            this->priv.alen = this->priv.slen/4;
          else
            this->priv.alen = this->priv.slen/2;
          this->priv.blen = this->priv.slen/4;  // always give the small size a shot
          }
        break;
        }
      /*@\\\0000000A1B+9543*/
      /*@/// case 18: */
      case 18: {
        if (this->calc_opt.fast) {
          this->priv.alen = this->priv.slen/2;
          this->priv.blen = this->priv.slen/2;
          }
        else {
          this->priv.alen = this->priv.slen/8;  // never require the multiword hash
          if (this->priv.nkeys <= this->priv.slen*(5.0/8.0))
            this->priv.blen = this->priv.slen/4;
          else
            this->priv.blen = this->priv.slen/2;
          }
        break;
        }
      /*@\\\000000021B+905F*/
      /*@/// case 19..20: */
      case 19:
      case 20: {
        if (this->priv.nkeys <= this->priv.slen*(5.0/8.0)) {
          this->priv.alen = this->priv.slen/8;
          this->priv.blen = this->priv.slen/4;
          }
        else {
          this->priv.alen = this->priv.slen/2;
          this->priv.blen = this->priv.slen/2;
          }
        break;
        }
      /*@\\\0000000206+A676*/
      /*@/// else */
      default: {
        this->priv.alen = this->priv.slen/2;  // just find a hash as quick as possible
        this->priv.blen = this->priv.slen/2;  // we'll be thrashing virtual memory at this size
        break;
        }
      /*@\\\+0379*/
      }
    /*@\\\0000000207+8E8F*/
    }
  else {
    /*@/// normal */
    if ( (this->priv.nkeys >= 1024) &&  // parameter?
         ( (this->priv.nkeys > this->priv.slen*this->calc_opt.keyspace_factor) ||
           ( (this->calc_opt.fast) &&
             (this->priv.nkeys > this->priv.slen*0.8) ))) {
      this->priv.slen = this->priv.slen * 2;
      }

    if (this->priv.slen < UB4BITS)
      this->priv.blen = this->priv.slen;  // go for function speed not space
    else if (this->priv.slen/4 <= 1 << 14) {
      if (this->priv.nkeys <= this->priv.slen*0.56)
        this->priv.blen = this->priv.slen/32;
      else if (this->priv.nkeys <= this->priv.slen*0.74)
        this->priv.blen = this->priv.slen/16;
      else
        this->priv.blen = this->priv.slen/8;
      }
    else {
      if (this->priv.nkeys <= this->priv.slen*0.6)
        this->priv.blen = this->priv.slen/16;
      else if (this->priv.nkeys <= this->priv.slen*0.8)
        this->priv.blen = this->priv.slen/8;
      else
        this->priv.blen = this->priv.slen/4;
      }

    if ((this->calc_opt.fast) &&
        (this->priv.blen < this->priv.slen/8))
      this->priv.blen = this->priv.slen/8;


    if (this->priv.slen > 131072) {
      // blen must be defined! BUG 3 in Jenkins' code
      this->priv.alen = ((t_32u)(1) << (UB4BITS-ceil_log2(this->priv.blen)));
        // distinct keys => distinct (A,B)
      }
    else
      this->priv.alen = this->priv.slen;  // no reason to restrict alen to slen/2

    if (this->priv.alen < 1)
      this->priv.alen = 1;
    if (this->priv.blen < 1)
      this->priv.blen = 1;
    /*@\\\+4EC1*/
    }
  }
/*@\\\0000001001+202B*/
/*@/// mycall void to_perfect_hash::scrambleinit() */
mycall void to_perfect_hash::scrambleinit(
  ) {
// initialize p_scramble[] with distinct random values in 0..slen-1

  t_int i;
  t_32u log_smax;
  t_32u limit;

  /*@/// realloc this->priv.p_scramble */
  if (this->priv.blen < this->calc_opt.use_scramble)  // see augment
    limit = this->priv.slen;  // should be <>0
  else
    limit = 255+1;

  if ((this->priv.p_scramble != nil) &&
      (this->priv.allocated_scramble == limit))
    return;  // no change!
  if (this->priv.p_scramble != nil) {
    free(this->priv.p_scramble);
    this->priv.p_scramble = nil;
    }
  this->priv.p_scramble =
    (t_32u*)calloc(limit, sizeof(this->priv.p_scramble[0]));
  this->priv.allocated_scramble = limit;
  /*@\\\000000060C+AB44*/

  // fill p_scramble[] with distinct random integers in 0..slen-1
  log_smax = ceil_log2(this->priv.slen);
  for (i = 0; i < (t_int)(limit); ++i) {
    this->priv.p_scramble[i] = permute(i, log_smax);
    }
  }
/*@\\\0000000801+2E24*/
/*@/// mycall void to_perfect_hash::setlow() */
mycall void to_perfect_hash::setlow(
  ) {
// find the highest and lowest bit where any key differs

  t_int i;
  tpr_key mykey;
  t_32u firstkey;
  t_32u my_diffbits;

  // mark the interesting bits in this->state_engine.mask
  my_diffbits = 0;  // copy of this->priv.diffbits
  mykey = this->priv.keys;
  firstkey = mykey->key_k;  // mykey can not be nil here, see generate_hash()
  while (mykey != nil) {
    my_diffbits = my_diffbits | (firstkey ^ mykey->key_k);
    if (my_diffbits == (t_32u)(-1))
      break;  // early exit
    mykey = mykey->next_k;
    }  // while mykey != nil

  // find the lowest interesting bit
  for (i = 0; i <= UB4BITS-1; ++i) {
    // if ((my_diffbits & ((t_32u)(1) << i)) != 0)
    if (((my_diffbits >> i) & 1) != 0)
      break;  // for i
    }  // for i
  this->priv.lowbit = i;

  // find the highest interesting bit
  for (i = UB4BITS-1; i >= 0; --i) {
    // if ((my_diffbits & ((t_32u)(1) << i)) != 0)
    if (((my_diffbits >> i) & 1) != 0)
      break;  // for i
    }  // for i
  this->priv.highbit = i;
  this->priv.diffbits = my_diffbits;
  }
/*@\\\0000000105+12A5*/
/*@/// mycall void to_perfect_hash::inithex(...) */
mycall void to_perfect_hash::inithex(
  t_32u salt  // used to initialize the hash function
  ) {
/*@/// doc */
/*
 * Initialize (a,b) when keys are integers.
 *
 * Normally there's an initial hash which produces a number.  That hash takes
 * an initializer.  Changing the initializer causes the initial hash to
 * produce a different (uniformly distributed) number without any extra work.
 *
 * Well, here we start with a number.  There's no initial hash.  Any mixing
 * costs extra work.  So we go through a lot of special cases to minimize the
 * mixing needed to get distinct (a,b).  For small sets of keys, it's often
 * fastest to skip the final hash and produce the perfect hash from the number
 * directly.
 *
 * The target user for this is switch statement optimization.  The common case
 * is 3 to 16 keys, and instruction counts matter.  The competition is a
 * binary tree of branches.
 *
 * Return TRUE if we found a perfect hash and no more work is needed.
 * Return FALSE if we just did an initial hash and more work is needed.
 */
/*@\\\+00CE*/
/*@/// doc */
/*
 * Guns aren't enough.  Bring out the Bomb.  Use tab[].
 * This finds the initial (a,b) when we need to use tab[].
 *
 * We need to produce a different (a,b) every time this is called.  Try all
 * reasonable cases, fastest first.
 *
 * The initial mix (which this determines) can be filled into final starting
 * at line[1].  val is set and a,b are declared.  The final hash (at line[7])
 * is a^tab[b] or a^p_scramble[tab[b]].
 *
 * The code will probably look like this, minus some stuff:
 *     val += CONSTANT;
 *     val ^= (val<<16);
 *     val += (val>>8);
 *     val ^= (val<<4);
 *     b = (val >> j) & 7;
 *     a = (val + (val<<k)) >> 29;
 *     return a^p_scramble[tab[b]];
 * Note that *a* and tab[b] will be computed in parallel by most modern chips.
 *
 * this->state_engine.i is the current state of the state machine.
 * this->state_engine.j and k are counters in the loops the states simulate.
 */
/*@\\\+00CE*/

  t_32u alog;
  t_32u blog;
  t_32u addk;

  if (this->priv.diffbits == 0)  // optimized to be calculated only when needed
    this->setlow();

  if (salt == 1) {
    this->state_engine.i = 1;
    this->state_engine.j = 0;
    this->state_engine.k = 0;
    }

  alog = ceil_log2(this->priv.alen);
  blog = ceil_log2(this->priv.blen);
  while (true) {
    switch (this->state_engine.i) {
      /*@/// case 1:   < HM_AB_1 */
      case 1: {
        // a = val>>30;  b=val&3
        if (! assigned(this->cb.a_gen[HM_AB_1])) {
          ++this->state_engine.i;
          continue;  // while true
          }

        this->gen_opt.hash_method = HM_AB_1;
        this->gen_opt.a_shl = UB4BITS-1-this->priv.highbit;
        this->gen_opt.a_shr = UB4BITS-alog;
        this->gen_opt.b_shr = this->priv.lowbit;
        this->gen_opt.b_mask = this->priv.blen-1;

        ++this->state_engine.i;
        return;
        break;
        }
      /*@\\\+A33A*/
      /*@/// case 2:   < HM_AB_2 */
      case 2: {
        // a = val&3;  b=val>>30
        if (! assigned(this->cb.a_gen[HM_AB_2])) {
          ++this->state_engine.i;
          continue;  // while true
          }

        this->gen_opt.hash_method = HM_AB_2;
        this->gen_opt.a_shr = this->priv.lowbit;
        this->gen_opt.a_mask = this->priv.alen-1;
        this->gen_opt.b_shl = UB4BITS-1-this->priv.highbit;
        this->gen_opt.b_shr = UB4BITS-blog;

        ++this->state_engine.i;
        return;
        break;
        }
      /*@\\\+B2C4*/
      /*@/// case 3:   < HM_AB_1 */
      case 3: {
        /*
         * cases 3,4,5:
         * for (k=lowbit; k<=highbit; ++k)
         *   for (j=lowbit; j<=highbit; ++j)
         *     b = (val>>j)&3;
         *     a = (val<<k)>>30;
         */
        if (! assigned(this->cb.a_gen[HM_AB_1])) {
          this->state_engine.i = 6;
          continue;  // while true
          }

        this->state_engine.k = this->priv.lowbit;
        this->state_engine.j = this->priv.lowbit;
        ++this->state_engine.i;
        break;
        }
      /*@\\\+C24C*/
      /*@/// case 4:   < " */
      case 4: {
        if (this->state_engine.j >= this->priv.highbit) {
          ++this->state_engine.i;
          continue;  // while true
          }

        this->gen_opt.hash_method = HM_AB_1;
        this->gen_opt.a_shl = UB4BITS-1-this->state_engine.k;
        this->gen_opt.a_shr = UB4BITS-alog;
        this->gen_opt.b_shr = this->state_engine.j;
        this->gen_opt.b_mask = this->priv.blen-1;

        ++this->state_engine.j;
        while (this->state_engine.j < this->priv.highbit) {
          if (( (this->priv.diffbits >> (this->state_engine.j)) &
                (this->priv.blen-1) ) > 2)
            break;  // while
          ++this->state_engine.j;
          }  // while
        return;
        break;
        }
      /*@\\\+3598*/
      /*@/// case 5:     " */
      case 5: {
        ++this->state_engine.k;
        while (this->state_engine.k < this->priv.highbit) {
          if (( (( this->priv.diffbits <<
                   (UB4BITS-1-this->state_engine.k) ) >> alog) &
                (this->priv.alen-1) ) > 0)
            break;  // while
          ++this->state_engine.k;
          }
        if (this->state_engine.k >= this->priv.highbit) {
          ++this->state_engine.i;
          continue;  // while true
          }
        this->state_engine.j = this->priv.lowbit;
        this->state_engine.i = 4;
        break;
        }
      /*@\\\+8E15*/
      /*@/// case 6:     HM_AB_3 */
      case 6: {
        /*
         * cases 6,7,8:
         * for (k=0; k<UB4BITS-alog; ++k)
         *   for (j=0; j<UB4BITS-blog; ++j)
         *     val = val+f(salt);
         *     val ^= (val >> 16);
         *     val += (val << 8);
         *     val ^= (val >> 4);
         *     b = (val >> j) & 3;
         *     a = (val + (val << k)) >> 30;
         */
        if (! assigned(this->cb.a_gen[HM_AB_3])) {
          this->state_engine.i = 9;
          continue;  // while true
          }

        this->state_engine.k = 0;
        this->state_engine.j = 0;
        ++this->state_engine.i;
        break;
        }
      /*@\\\+29CD*/
      /*@/// case 7:   < " */
      case 7: {
        // Just do something that will surely work
        addk = 0x9e3779b9*salt;

        if (this->state_engine.j > UB4BITS-blog) {
          ++this->state_engine.i;
          continue;  // while true
          }

        this->gen_opt.hash_method = HM_AB_3;
        this->gen_opt.salt = addk;
        this->gen_opt.do_xor_shr_16 =
          (this->priv.highbit - this->priv.lowbit > 16-1);
        this->gen_opt.do_add_shl_8 =
          (this->priv.highbit - this->priv.lowbit > 8-1);
        this->gen_opt.a_shl = this->state_engine.k;
        this->gen_opt.a_shr = UB4BITS-alog;
        this->gen_opt.b_shr = this->state_engine.j;
        this->gen_opt.b_mask = this->priv.blen-1;

        ++this->state_engine.j;
        return;
        break;
        }
      /*@\\\+B6A3*/
      /*@/// case 8:     " */
      case 8: {
        ++this->state_engine.k;
        if (this->state_engine.k > UB4BITS-alog) {
          ++this->state_engine.i;
          continue;  // while true
          }

        this->state_engine.j = 0;
        this->state_engine.i = 7;
        break;
        }
      /*@\\\+0F20*/
      /*@/// case 9:     HM_AB_X1 */
      case 9: {
        if (! assigned(this->cb.a_gen[HM_AB_X1])) {
          this->state_engine.i = 11;
          continue;  // while true
          }

        this->state_engine.k = 0;
        this->state_engine.j = 0;
        ++this->state_engine.i;
        break;
        }
      /*@\\\+18C6*/
      /*@/// case 10:  < " */
      case 10: {
        addk = (SEARCH_START+salt*SEARCH_OFFSET) | 1;

        if (this->state_engine.j > (UB4BITS-alog)*(UB4BITS-blog)) {
          // same count as method 6/7/8
          ++this->state_engine.i;
          continue;  // while true
          }

        if (alog+blog+this->priv.lowbit > UB4BITS) {
          ++this->state_engine.i;
          continue;  // while true
          }

        this->gen_opt.hash_method = HM_AB_X1;
        this->gen_opt.salt = addk;
        this->gen_opt.a_shr = UB4BITS-(alog+blog+this->priv.lowbit);
        this->gen_opt.a_mask = this->priv.alen-1;
        this->gen_opt.b_shr = UB4BITS-blog;
        this->gen_opt.b_mask = this->priv.blen-1;

        ++this->state_engine.j;
        return;
        break;
        }
      /*@\\\0000000201+730D*/
      /*@/// case 11:    HM_AB_X2 */
      case 11: {
        if (! assigned(this->cb.a_gen[HM_AB_X2])) {
          this->state_engine.i = 13;
          continue;  // while true
          }

        this->state_engine.k = 0;
        this->state_engine.j = 0;
        ++this->state_engine.i;
        break;
        }
      /*@\\\+6B48*/
      /*@/// case 12:  < " */
      case 12: {
        addk = SEARCH_START+salt*SEARCH_OFFSET;

        if (this->state_engine.j > (UB4BITS-alog)*(UB4BITS-blog)) {
          // same count as method 6/7/8
          ++this->state_engine.i;
          continue;  // while true
          }

        this->gen_opt.hash_method = HM_AB_X2;
        this->gen_opt.salt = addk;
        this->gen_opt.a_shr = (UB4BITS-alog-blog) & (UB4BITS-1);
        this->gen_opt.a_mask = this->priv.alen-1;
        this->gen_opt.b_shr = UB4BITS-blog;
        this->gen_opt.b_mask = this->priv.blen-1;

        ++this->state_engine.j;
        return;
        break;
        }
      /*@\\\000000010F+B86F*/
      /*@/// case 13:    HM_AB_X3 */
      case 13: {
        if (! assigned(this->cb.a_gen[HM_AB_X3])) {
          this->state_engine.i = 15;
          continue;  // while true
          }

        this->state_engine.k = 0;
        this->state_engine.j = 0;
        ++this->state_engine.i;
        break;
        }
      /*@\\\+8A5B*/
      /*@/// case 14:  < " */
      case 14: {
        addk = SEARCH_START+salt*SEARCH_OFFSET;

        if (this->state_engine.j > (UB4BITS-alog)*(UB4BITS-blog)) {
          // same count as method 6/7/8
          ++this->state_engine.i;
          continue;  // while true
          }

        this->gen_opt.hash_method = HM_AB_X3;
        this->gen_opt.salt = addk;
        this->gen_opt.a_mask = this->priv.alen-1;
        this->gen_opt.b_shr = UB4BITS-blog;

        ++this->state_engine.j;
        return;
        break;
        }
      /*@\\\+FDD1*/
      /*@/// case 15:    - */
      case 15: {
        if (assigned(this->cb.a_gen[HM_AB_3]))
          this->state_engine.i = 6;
        else if (assigned(this->cb.a_gen[HM_AB_X1]) ||
                 assigned(this->cb.a_gen[HM_AB_X2]) ||
                 assigned(this->cb.a_gen[HM_AB_X3]))
          this->state_engine.i = 9;
        else
          this->state_engine.i = 1;
        break;
        }
      /*@\\\0000000227+78B1*/
      }
    }
  }
/*@\\\0000002101+F10A*/
/*@/// mycall tq_reason to_perfect_hash::inittab() */
mycall tq_reason to_perfect_hash::inittab(
  ) {
/*@/// doc */
/*
 * Run a hash function on the key to get a and b
 * Returns:
    REASON_SUCCESS,  // found distinct (a,b) for all keys, put keys in tabb[]
    REASON_FAILURE,  // didn't find distinct (a,b) for all keys
    REASON_LIMIT,  // limit reached, early abort
    REASON_DUPE);  // a real dupe found
 */
/*@\\\0000000801+00CE*/
/*
 * put keys in tabb according to tr_key->b_k
 * check if (the initial hash might work
 */

  tpr_key mykey;
  tpr_key otherkey;
  tpr_bstuff p_bstuff;
  tf_calc_ab f_calc_ab;

  f_calc_ab = a_f_calc_ab[this->gen_opt.hash_method];
  ZERO_MEM(*this->priv.p_tabb, sizeof(this->priv.p_tabb[0])*this->priv.blen);

  // Two keys with the same (a,b) guarantees a collision
  mykey = this->priv.keys;
  while (mykey != nil) {
    f_calc_ab(this->gen_opt, *mykey);
    p_bstuff = &this->priv.p_tabb[mykey->b_k];
    otherkey = p_bstuff->list_b;
    while (otherkey != nil) {
      if (mykey->a_k == otherkey->a_k) {
        // this->checkdup(mykey, otherkey);
        if (mykey->key_k == otherkey->key_k) {
          return REASON_DUPE;
          }

        return REASON_FAILURE;
        }
      otherkey = otherkey->nextb_k;
      }

    if (p_bstuff->listlen_b > this->calc_opt.listlen_limit) {
      return REASON_LIMIT;  // speedup
      }

    ++p_bstuff->listlen_b;
    mykey->nextb_k = p_bstuff->list_b;
    p_bstuff->list_b = mykey;
    mykey = mykey->next_k;
    }

  return REASON_SUCCESS;
  }
/*@\\\0000000104+FA14*/
/*@/// mycall t_bool to_perfect_hash::apply(...) */
mycall t_bool to_perfect_hash::apply(
  tr_hstuff *tabh,  // arry
  tr_qstuff *tabq,  // array
  t_32u tail,
  t_bool rollback  // FALSE applies augmenting path, TRUE rolls back
  ) {
// Try to apply an augmenting list

  t_32u hash;
  tpr_key mykey;
  tpr_bstuff pb;
  t_32u child;
  t_32u parent;
  t_32u stabb;  // p_scramble[tab[b]]
  t_32u *my_p_scramble;  // array

  my_p_scramble = this->priv.p_scramble;  // local copy to speed up
  // walk from child to parent
  child = tail-1;
  while (child != 0) {
    parent = tabq[child].parent_q;  // find child's parent
    pb     = tabq[parent].b_q;  // find parent's list of siblings

    // erase old hash values
    stabb = my_p_scramble[pb->val_b];
    mykey = pb->list_b;
    while (mykey != nil) {
      hash = mykey->a_k ^ stabb;
      if (mykey == tabh[hash].key_h) {
        // erase hash for all of child's siblings
        tabh[hash].key_h = nil;
        }
      mykey = mykey->nextb_k;
      }  // while mykey != nil

    // change pb->val_b, which will change the hashes of all parent siblings
    if (rollback)
      pb->val_b = tabq[child].oldval_q;
    else
      pb->val_b = tabq[child].newval_q;

    // set new hash values
    stabb = my_p_scramble[pb->val_b];
    mykey = pb->list_b;
    while (mykey != nil) {
      hash = mykey->a_k ^ stabb;
      if (rollback) {
        if (parent == 0)
          goto CONT;  // continue while, root never had a hash
        }
      else if (tabh[hash].key_h != nil) {
        // very rare: roll back any changes
        this->apply(tabh, tabq, tail, true);  // recusive call
        return false;  // failure, collision
        }
      tabh[hash].key_h = mykey;
CONT:
      mykey = mykey->nextb_k;
      }  // while mykey != nil
    child = parent;
    }  // while child != 0
  return true;
  }
/*@\\\0000003F04+6934*/
/*@/// mycall tq_reason to_perfect_hash::augment(...) */
mycall tq_reason to_perfect_hash::augment(
  tr_hstuff *tabh,  // array; which key is associated with which hash, indexed by hash
  tr_qstuff *tabq,  // array; queue of *b* values, this is the spanning tree
  tpr_bstuff item,  // &tabb[b] for the b to be mapped
  t_32u highwater   // a value higher than any now in tabb[].water_b
  ) {
/*@/// doc */
/*
-------------------------------------------------------------------------------
augment(): Add item to the mapping.

Construct a spanning tree of *b*s with *item* as root, where each
parent can have all its hashes changed (by some new val_b) with
at most one collision, and each child is the b of that collision.

I got this from Tarjan's "Data Structures and Network Algorithms".  The
path from *item* to a *b* that can be remapped with no collision is
an "augmenting path".  Change values of tab[b] along the path so that
the unmapped key gets mapped and the unused hash value gets used.

Assuming 1 key per b, if m out of n hash values are still unused,
you should expect the transitive closure to cover n/m nodes before
an unused node is found.  Sum(i=1..n)(n/i) is about nlogn, so expect
this approach to take about nlogn time to map all single-key b's.
-------------------------------------------------------------------------------
*/
/*@\\\+00CE*/
/*@/// var */

  t_32u q;                    // current position walking through the queue
  t_32u tail;            // tail of the queue.  0 is the head of the queue.
  t_32u limit;

  tpr_bstuff myb;                                    // the b for this node
  t_32u i;                                 // possible value for myb->val_b

  tpr_bstuff childb;                           // the b that this i maps to
  tpr_key mykey;                          // for walking through myb's keys

  tpr_key childkey;
  t_32u hash;

  tpr_bstuff hitb;
  tpr_qstuff my_qstuff;
  t_32u my_scramble;
  tr_bstuff *my_p_tabb;  // array
  t_32u *my_p_scramble;  // array
  t_32u my_highhash;
  t_32u my_count_limit;
  t_bool my_trans;
  t_32u my_count;
/*@\\\+40D1*/
  tq_reason res;

  my_p_tabb = this->priv.p_tabb;  // local copy to speed up
  my_p_scramble = this->priv.p_scramble;  // local copy to speed up
  my_highhash = this->priv.highhash;  // local copy to speed up
  my_count_limit = this->calc_opt.count_limit;  // local copy to speed up
  my_trans = this->priv.trans;  // local copy to speed up

  if (this->priv.blen < this->calc_opt.use_scramble)
    limit = this->priv.slen;  // should be  != 0
  else
    limit = 255+1;

  // initialize the root of the spanning tree
  tabq[0].b_q = item;
  tail = 1;

  my_count = 0;
  // construct the spanning tree by walking the queue, add children to tail
  q = 0;
  do {  // tail changes in loop
    /*@/// loop q */
    myb = tabq[q].b_q;  // the b for this node

    for (i = 0; i < limit; ++i) {  // possible value for myb->val_b
      ++my_count;
      if (my_count > my_count_limit) {
        res = REASON_LIMIT;
        goto STOP;  // already too much work for this key, give up
        }
      /*@/// loop i */
      my_scramble = my_p_scramble[i];
      childb = nil;  // the b that this i maps to

      mykey = myb->list_b;  // for walking through myb's keys
      while (mykey != nil) {
        /*@/// while mykey */
        hash = mykey->a_k ^  my_scramble;

        if (hash >= my_highhash) {
          // break;  // while mykey, out of bounds
          goto CONT;  // continue for i
          // should trigger only for (near) minimal hash
          }
        childkey = tabh[hash].key_h;

        if (childkey != nil) {
          hitb = &my_p_tabb[childkey->b_k];

          if (childb == nil) {
            childb = hitb;  // remember this as childb
            if (hitb->water_b == highwater) {  // hitb=childb
              // break;  // while mykey, already explored
              goto CONT;  // continue for i
              }
            }
          else {
            if (childb != hitb) {
              // break;  // while mykey, hit at most one child b
              goto CONT;  // continue for i
              }
            }
          }
        /*@\\\0000000115+DE38*/
        mykey = mykey->nextb_k;
        }  // while mykey

      // if (mykey != nil)
      //   continue;  // for i, myb with i has multiple collisions

      // add childb to the queue of reachable things
      my_qstuff = &tabq[tail];
      my_qstuff->b_q = childb;
      my_qstuff->oldval_q = myb->val_b;  // need this for rollback
      my_qstuff->newval_q = i;   // how to make parent (myb) use this hash
      my_qstuff->parent_q = q;
      ++tail;

      if (childb != nil)
        childb->water_b = highwater;
      else {
        // found an *i* with no collisions?
        // try to apply the augmenting path
        if (this->apply(tabh, tabq, tail, false)) {
          res = REASON_SUCCESS;
          // EXIT;  // success, item was added to the perfect hash
          goto STOP;  // success
          }

        --tail;  // don't know how to handle such a child!
        }
      /*@\\\0000000601+4DCB*/
    CONT: {}
      }  // for i
    /*@\\\0000000901+4DD3*/
    ++q;
    } while ( (my_trans) &&  // don't do transitive closure, i.e. give up
              (q < tail) );  // do..while q
  res = REASON_FAILURE;  // while q exceeded, i.e. could not augment
STOP:
  this->gen_opt.count_for_limit += my_count;
  // finalize statistics
  if (my_count > this->gen_opt.max_count)
    this->gen_opt.max_count = my_count;
  return res;
  }
/*@\\\000000202A+D264*/
/*@/// mycall tq_reason to_perfect_hash::perfect(...) */
mycall tq_reason to_perfect_hash::perfect(
  tr_hstuff *tabh,  // tar_hstuff
  tr_qstuff *tabq  // tar_qstuff
  ) {
// find a mapping that makes this a perfect hash

  t_32u maxkeys;  // maximum number of keys for any b
  t_32u i, j;
  tr_bstuff *my_p_tabb;  // array
  tq_reason res;

  // clear any state from previous attempts
  // assume this->priv.blen != 0
  my_p_tabb = this->priv.p_tabb;  // local copy to speed up
  ZERO_MEM(*tabh, sizeof(tabh[0])*this->priv.slen);  // BUG 1 in Jenkins' code
  ZERO_MEM(*tabq, sizeof(tabq[0])*(this->priv.blen+1));

  maxkeys = 0;
  for (i = 0; i < this->priv.blen; ++i) {
    if (my_p_tabb[i].listlen_b > maxkeys)
      maxkeys = my_p_tabb[i].listlen_b;
    }

  TRACE("maxkeys="+num(maxkeys));

  // In descending order by number of keys, map all *b*s
  for (j = maxkeys; j >= 1; --j) {
    for (i = 0; i < this->priv.blen; ++i) {
      if (my_p_tabb[i].listlen_b == j) {
        res = this->augment(tabh, tabq, &my_p_tabb[i], i+1);
        if (res != REASON_SUCCESS) {
          TRACE(
            "fail to map group of size "+num(j)+" for tab size "+
            num(this->priv.blen));
          return res;
          }
        }
      }
    }

  // Success!  We found a perfect hash of all keys into 0..nkeys-1.
  return REASON_SUCCESS;
  }
/*@\\\+B48C*/
/*@/// mycall void to_perfect_hash::finalize_perfect() */
mycall void to_perfect_hash::finalize_perfect(
  ) {
// Optimized to get rid of tabb containing only 0 values
// fill gen_opt

  t_32u i;
  t_bool zero;

  /*@/// kill tabb if all zero */
  if (this->priv.blen > 0) {
    zero = true;
    for (i = 0; i < this->priv.blen; ++i) {
      if (this->priv.p_tabb[i].val_b != 0) {
        zero = false;
        break;  // for i
        }
      }  // for i
    if (zero) {
      TRACE("tabb not needed");
      if (this->priv.p_tabb != nil) {
        free(this->priv.p_tabb);
        this->priv.p_tabb = nil;
        }
      this->priv.blen = 0;
      }
    }
  /*@\\\0000000A05+C82D*/
  /*@/// fill gen_opt */
  if (this->gen_opt.p_tabb != nil) {
    free(this->gen_opt.p_tabb);
    this->gen_opt.p_tabb = nil;
    }
  this->gen_opt.p_scramble = this->priv.p_scramble;
  this->gen_opt.scramble_len = 0;
  this->gen_opt.scramble_max = 0;
  this->gen_opt.tabb_len = this->priv.blen;
  this->gen_opt.tabb_max = 0;
  if (this->priv.blen > 0) {
    if (false) {}
    else if ((this->priv.slen <= 255+1) ||
            (this->priv.blen < this->calc_opt.use_scramble)) {
      this->gen_opt.tabb_max = this->priv.slen-1;
      this->gen_opt.p_tabb = (t_32u*)
        calloc(this->priv.blen, sizeof(this->gen_opt.p_tabb[0]));
      for (i = 0; i < this->priv.blen; ++i) {
        this->gen_opt.p_tabb[i] =
          this->priv.p_scramble[this->priv.p_tabb[i].val_b];
        }
      }
    else {
      this->gen_opt.tabb_max = 255;
      this->gen_opt.scramble_len = 256;
      this->gen_opt.scramble_max = this->priv.slen-1;
      this->gen_opt.p_tabb = (t_32u*)
        calloc(this->priv.blen, sizeof(this->gen_opt.p_tabb[0]));
      for (i = 0; i < this->priv.blen; ++i)
        this->gen_opt.p_tabb[i] = this->priv.p_tabb[i].val_b;
      }
    }
  // this->gen_opt.hash_used_max => calc_max
  this->calc_max();

  this->gen_opt.hash_gen_max = this->priv.slen-1;
  this->gen_opt.check_value = true;
  this->gen_opt.check_range = false;
  /*@\\\+CB63*/
  }
/*@\\\0000000901+AAC9*/
/*@/// mycall t_bool to_perfect_hash::findhash() */
mycall t_bool to_perfect_hash::findhash(
  ) {
/*@/// doc */
/*
  ** Try to find a perfect hash function.
  ** Return the successful initializer for the initial hash.
  ** Return 0 if (no perfect hash could be found.
*/
/*@\\\0000000103+00CE*/

  t_32u bad_initkey;  // how many times did initkey() fail?
  t_32u bad_perfect;  // how many times did perfect() fail?
  t_32u salt;  // trial initializer for initial hash
  t_32u maxalen;
  tr_hstuff *p_tabh;  // array; table of keys indexed by hash value
  tr_qstuff *p_tabq;  // array; table of stuff indexed by queue value,
                       // used by augment()
  tq_reason rslinit;
  t_bool res;

  this->priv.diffbits = 0;

  if (! assigned(this->cb.a_gen[HM_AB_1]) &&
      ! assigned(this->cb.a_gen[HM_AB_2]) &&
      ! assigned(this->cb.a_gen[HM_AB_3]) &&
      ! assigned(this->cb.a_gen[HM_AB_X1]) &&
      ! assigned(this->cb.a_gen[HM_AB_X2]) &&
      ! assigned(this->cb.a_gen[HM_AB_X3])) {
    return false;
    }

  this->gen_opt.count_for_limit = 0;
  this->gen_opt.max_count = 0;

  p_tabh = nil;
  p_tabq = nil;
  // try

    // guess initial values for slen, alen and blen
    this->initalen();

    this->scrambleinit();

    if (this->calc_opt.minimal)
      maxalen = this->priv.slen / 2;
    else
      maxalen = this->priv.slen;

    if (this->calc_opt.minimal)
      this->priv.highhash = this->priv.nkeys;
    else {
      // this->priv.highhash = this->priv.slen;
      this->priv.highhash = min_32(
        this->priv.slen,
        (t_32u)(this->priv.nkeys * this->calc_opt.near_minimal_factor));
        // round
      }
    this->priv.trans =
      (! this->calc_opt.fast) ||
      (this->calc_opt.fast_trans) ||
      (this->calc_opt.minimal) ||
      (this->calc_opt.near_minimal_factor < 2.0);

    // allocate working memory
    /*@/// realloc this->priv.p_tabb */
    if (this->priv.p_tabb != nil) {
      free(this->priv.p_tabb);
      this->priv.p_tabb = nil;
      }
    this->priv.p_tabb = (tr_bstuff*)calloc(this->priv.blen, sizeof(tr_bstuff));
    /*@\\\0000000527+F0D4*/
    p_tabq = (tr_qstuff*)calloc(this->priv.blen+1, sizeof(tr_qstuff));
    p_tabh = (tr_hstuff*)calloc(this->priv.slen, sizeof(tr_hstuff));
      // BUG 1 in Jenkins' code

    // Actually find the perfect hash
    bad_initkey = 0;
    bad_perfect = 0;
    salt = 1;
    TRACE("Using "+num(this->priv.alen)+" / "+num(this->priv.blen));
    while (true) {
      /*@/// Try to find distinct (A,B) for all keys */
      // Try to find distinct (A,B) for all keys
      TRACE_ODO(num(salt));

      // the initial hash of the keys
      this->inithex(salt);
      rslinit = this->inittab();
      switch (rslinit) {
        /*@/// case REASON_FAILURE,REASON_LIMIT: didn't find distinct (a,b) */
        case REASON_FAILURE:
        case REASON_LIMIT: {
          // didn't find distinct (a,b)
          ++bad_initkey;
          if (bad_initkey >= this->calc_opt.retry_initkey) {
            // Try to put more bits in (A,B) to make distinct (A,B) more likely
            if (this->priv.alen < maxalen) {
              this->priv.alen = this->priv.alen*2;
              }
            else if (this->priv.blen < this->priv.slen) {
              this->priv.blen = this->priv.blen*2;
              /*@/// realloc p_tabq */
              if (p_tabq != nil) {
                free(p_tabq);
                p_tabq = nil;
                }
              p_tabq = (tr_qstuff*)calloc(this->priv.blen+1, sizeof(tr_qstuff));
              /*@\\\0000000530+3869*/
              /*@/// realloc this->priv.p_tabb */
              if (this->priv.p_tabb != nil) {
                free(this->priv.p_tabb);
                this->priv.p_tabb = nil;
                }
              this->priv.p_tabb =
                (tpr_bstuff)calloc(this->priv.blen, sizeof(tr_bstuff));
              /*@\\\0000000539+F0D4*/
              this->scrambleinit();
              }
            else {
              TRACE("fatal error: Cannot perfect hash: cannot find distinct (A,B)");
              res = false;
              goto EXIT;
              }
            TRACE(
              num(salt)+": enlarging table to "+num(this->priv.alen)+" / "+
              num(this->priv.blen));
            bad_initkey = 0;
            bad_perfect = 0;
            this->state_engine.i = 1;  // try simpler versions first
            }
          goto CONT;  // two keys have same (a,b) pair
          break;
          }
        /*@\\\0000001A3D+2144*/
        /*@/// case REASON_DUPE: a real dupe found */
        case REASON_DUPE: {
          res = false;
          goto EXIT;
          break;
          }
        /*@\\\0000000401+4D42*/
        case REASON_SUCCESS: break;
        }

      TRACE(num(salt)+": found distinct (A,B) ["+num(this->state_engine.i)+"]");

      // Given distinct (A,B) for all keys, build a perfect hash
      rslinit = this->perfect(p_tabh, p_tabq);
      if (rslinit != REASON_SUCCESS) {
        /*@/// no success */
        if (rslinit == REASON_LIMIT) {
          TRACE("max count reached");
          // bad_perfect += 10;
          }
        if ((this->calc_opt.work_limit != 0) &&
           (this->gen_opt.count_for_limit > this->calc_opt.work_limit)) {
          TRACE("max work reached");
          bad_perfect = this->calc_opt.retry_perfect;
          this->gen_opt.count_for_limit = 0;
          }

        ++bad_perfect;
        if (bad_perfect >= this->calc_opt.retry_perfect) {
          if ((this->priv.blen < this->priv.slen)) {
            this->priv.blen = this->priv.blen * 2;
            /*@/// realloc p_tabq */
            if (p_tabq != nil) {
              free(p_tabq);
              p_tabq = nil;
              }
            p_tabq = (tr_qstuff*)calloc(this->priv.blen+1, sizeof(tr_qstuff));
            /*@\\\0000000530+3869*/
            /*@/// realloc this->priv.p_tabb */
            if (this->priv.p_tabb != nil) {
              free(this->priv.p_tabb);
              this->priv.p_tabb = nil;
              }
            this->priv.p_tabb =
              (tpr_bstuff)calloc(this->priv.blen, sizeof(tr_bstuff));
            /*@\\\0000000539+F0D4*/
            scrambleinit();
            // --salt;  // we know this salt got distinct (A,B)
            this->state_engine.i = 1;  // better retry with simpler versions
            TRACE(
              num(salt)+": enlarging table to "+num(this->priv.alen)+" / "+
              num(this->priv.blen));
            }
          else {
            TRACE("fatal error: Cannot perfect hash: cannot build tab[]");
            res = false;
            goto EXIT;
            }
          bad_perfect = 0;
          }
        /*@\\\0000000E0E+E62E*/
        goto CONT;  // continue while true
        }

      // BREAK;  // while true
      goto DONE;  // break while true
      /*@\\\0000000301+A291*/
CONT:
      ++salt;
      }  // while true

DONE:
    TRACE(num(salt)+": built perfect hash table of size "+num(this->priv.blen));
    this->finalize_perfect();
    res = true;

EXIT: {}
  // finally
    // free working memory
    if (p_tabh != nil)
      free(p_tabh);
    if (p_tabq != nil)
      free(p_tabq);

  return res;
  }
/*@\\\0000003418+F101*/

/*@/// mycall void to_perfect_hash::calc_max() */
mycall void to_perfect_hash::calc_max(
  ) {

  t_32u max;
  t_32u v;
  tpr_key mykey;
  t_32u key_u;
  t_32u min_u;
  t_32u max_u;
  t_32s key_s;
  t_32s min_s;
  t_32s max_s;
  t_32u bits;
  t_bool use_signed;
  tf_calc_ab f_calc_ab;

  f_calc_ab = a_f_calc_ab[this->gen_opt.hash_method];
  max = 0;
  mykey = this->priv.keys;
  while (mykey != nil) {
    f_calc_ab(this->gen_opt, *mykey);
    v = calc_hash_tab(this->gen_opt, *mykey);
    if (v > max)
      max = v;
    mykey = mykey->next_k;
    }
  this->gen_opt.hash_used_max = max;
  this->gen_opt.is_minimal = (max+1 == this->priv.nkeys);

  this->priv.p_inv_hash = (tr_hstuff*)calloc(max+1, sizeof(tr_hstuff));
  mykey = this->priv.keys;
  while (mykey != nil) {
    v = calc_hash_tab(this->gen_opt, *mykey);
    this->priv.p_inv_hash[v].key_h = mykey;
    mykey = mykey->next_k;
    }

  mykey = this->priv.keys;
  key_u = mykey->key_k;
  min_u = key_u;
  max_u = key_u;
  key_s = (t_32s)(mykey->key_k);
  min_s = key_s;
  max_s = key_s;
  while (mykey != nil) {
    key_u = mykey->key_k;
    if (min_u > key_u)
      min_u = key_u;
    if (max_u < key_u)
      max_u = key_u;
    key_s = (t_32s)(mykey->key_k);
    if (min_s > key_s)
      min_s = key_s;
    if (max_s < key_s)
      max_s = key_s;
    mykey = mykey->next_k;
    }

  if (max_u <= 0xff) {
    bits = 8;
    use_signed = false;
    }
  else if ((-0x80 <= min_s) && (max_s <= 0x7f)) {
    bits = 8;
    use_signed = true;
    }
  else if (max_u <= 0xffff) {
    bits = 16;
    use_signed = false;
    }
  else if ((-0x8000 <= min_s) && (max_s <= 0x7fff)) {
    bits = 16;
    use_signed = true;
    }
  else {
    bits = 32;
    use_signed = false;
    }

  this->gen_opt.use_signed = use_signed;
  this->gen_opt.bits = bits;
  if (use_signed) {
    this->gen_opt.low = min_u;
    this->gen_opt.high = max_u;
    }
  else {
    this->gen_opt.low = (t_32u)(min_s);
    this->gen_opt.high = (t_32u)(max_s);
    }
  }
/*@\\\0000000E01+5E07*/
/*@\\\0000000201+0E34*/
/*@\\\0001000011001301*/
